#include "StdAfx.h"

#include "SimiyIni.h"
#include "SimiyUtil.h"
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

CSimiyIni	SimiyIni;

int IniCompare(const void *A, const void *B)
{
	IniNode *IniA = (IniNode *)A;
	IniNode *IniB = (IniNode *)B;

	if( _tcscmp( IniA->Section, IniB->Section ) || _tcscmp( IniA->Key, IniB->Key ) )
		return 1;	// different

	return 0;	// same section & key
}

CSimiyIni::CSimiyIni()
{
	m_Ini = new CLinkedList( sizeof( IniNode ), IniCompare );
	*m_IniPath = _TEXT('\0');
}

CSimiyIni::~CSimiyIni ()
{
	delete m_Ini;
}

void CSimiyIni::InitIni()
{
	for(IniNode *node = (IniNode *)m_Ini->dltofirst();
		node;
		node = (IniNode *)m_Ini->dlget() )
	{
		m_Ini->dldelete();
	}
}

bool CSimiyIni::ReadIni(LPCTSTR filepath)
{
	TCHAR buf[512];
	
	if( !FileExist(filepath) ) return false;

	FILE *inifile = _tfopen( filepath,  "rb");
	
	if(inifile == NULL) return false;

	this->InitIni();

	GetDirName(filepath, this->m_IniPath);

	IniNode node;
	TCHAR firstChar;
	TCHAR *token;
	
	while( _fgetts(buf, 511, inifile) )
	{
		firstChar = FirstLChar(buf);

		// comment or blank line
		if( firstChar == _T(';') || firstChar == _T('\0') )
			continue;
		// setting end
		if( ! _tcsncmp(buf, _T("_END_"), 5) )
			break;
		
		LTrim(buf);
		RTrim(buf);

		// section
		if( *buf == _T('[') && buf[ _tcslen(buf)-1 ] == _T(']') )
		{
			_tcsncpy(node.Section, buf+1, _tcslen(buf)-2);
			node.Section[_tcslen(buf)-2] = '\0';
			continue;
		}

		token = _tcstok(buf, _TEXT("=") );
		if( token == NULL ) continue;
		RTrim(token);
		_tcscpy(node.Key, token);

		token = _tcstok( NULL, _TEXT("\r\n") );
		if( token == NULL ) continue;
		LTrim(token);	// value start space trim
		_tcscpy(node.Value, token);

		m_Ini->dladd( (LPVOID) &node );

		//SimiyIni.SfvChkLogAppend("%s,%s,%s", node.Section, node.Key, node.Value);
	}
	fclose(inifile);

	return true;
}

int	CSimiyIni::GetValueInt(LPCTSTR section, LPCTSTR key, int Default)
{
	IniNode searchNode;
	_tcscpy( searchNode.Section, section );
	_tcscpy( searchNode.Key, key );

	IniNode *node = (IniNode *) m_Ini->dlfind( &searchNode );
	if( node )
	{
		return _ttoi(node->Value);
	}
	return Default;
}

LPCTSTR	CSimiyIni::GetValueString(LPCTSTR section, LPCTSTR key, LPCTSTR Default)
{
	IniNode searchNode;
	_tcscpy( searchNode.Section, section );
	_tcscpy( searchNode.Key, key );

	IniNode *node = (IniNode *) m_Ini->dlfind( &searchNode );
	if( node )
	{
		return node->Value;
	}
	return Default;
}

bool CSimiyIni::IsIncludeNoCntFExt(LPCTSTR filename)
{
	TCHAR NO_CNT_FEXT[256];
	TCHAR buf[512];

	_tcscpy(NO_CNT_FEXT, this->GetValueString("GENERAL","NO_CNT_FEXT"," "));

	GetExtName(filename, buf);

	TCHAR *token = _tcstok(NO_CNT_FEXT, " ");
	while( token )
	{
		if( ! _tcsicmp(buf, token) ) return true;

		token = _tcstok(NULL, " ");
	}
	return false;
}

bool CSimiyIni::SfvChkLogAppend(LPCTSTR format, ... )
{
	static TCHAR _buf[512];
	_stprintf(_buf, "%s%s", this->m_IniPath, SFVCHK_LOGFILE);

	FILE *fp = _tfopen(_buf, "ab");

	if(fp == NULL) return false;

	long ret;
	va_list args;
	va_start(args, format);
	ret = _vsntprintf(_buf, 512, format, args);
	va_end(args);

	time(&ret);
	struct tm *pTm = localtime( &ret );

	ret = _ftprintf(fp, "%04d/%02d/%02d %02d:%02d:%02d %s\r\n",
		pTm->tm_year + 1900, pTm->tm_mon + 1, pTm->tm_mday,
		pTm->tm_hour, pTm->tm_min, pTm->tm_sec,
		_buf);

	fflush(fp);	fclose(fp);

	if(ret < 0) return false;

	return true;
};