#include "stdafx.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "SimiyLog.h"
#include "SimiyIni.h"
#include "SimiyUtil.h"
#include "SimiySfvChk.h"
#include "SimiyCrc32.h"

#ifndef _tstol
	#define _tstol	_ttol
	#define _tstoi	_ttoi
#endif

CSimiyLog SimiyLog;			// Up Check Log File Processing Class

// RD2 [fileID]=[user]|[group]|[mod]|[isChecked]|[isFreeFile]|
// 

int LogCompare(const void *A, const void *B)
{
	double value = ((SNode *)A)->fileID - ((SNode *)B)->fileID;

	return ((value == 0) ? (0)
				: ((value < 0) ? (-1)
				: (1)));
}

CSimiyLog::CSimiyLog ()
{
	m_modified = false;
	m_Log = NULL;
	m_RaceUser	= NULL;
	m_RaceGroup	= NULL;
	m_Acl = NULL;
	m_Sfv = NULL;

	m_RacePeakFile = NULL;

	this->m_SfvCount = -1;
	this->m_SfvGood = 0;
	this->m_SfvUpFileCnt = 0;
};

CSimiyLog::~CSimiyLog ()
{
	if( m_Log )			delete m_Log;
	if( m_RaceUser )	delete m_RaceUser;
	if( m_RaceGroup )	delete m_RaceGroup;
	if( m_Acl )	delete m_Acl;
	if( m_Sfv )	delete m_Sfv;
}

void CSimiyLog::InitLog()
{
	SNode *node = (SNode *)m_Log->dltofirst();

	while( node )
	{
		m_Log->dldelete();
		node = (SNode *)m_Log->dlget();
	}
}


bool CSimiyLog::ReadLogFile(LPCTSTR filePath)
{
	TCHAR buf[128];
	
	// aleady readed
	if( m_Log ) return true;

	_tcscpy( this->m_LogPath, filePath );	// save file path

	m_Log = new CLinkedList( sizeof( SNode ), LogCompare );

	if( ! FileExist(filePath) ) return true;

	FILE *logfile = _tfopen( filePath,  "rb");
	
	if(logfile == NULL)
	{
		SimiyIni.SfvChkLogAppend( "SimiyLog::ReadLogFile() fopen failed - %s", strerror(errno) );
		delete m_Log;
		m_Log = NULL;
		return false;
	}

	while( _fgetts(buf, 121, logfile) )
	{
		SNode node;

		TCHAR *token = _tcstok( buf, _TEXT("|") );
		if( token == NULL ) continue;
		node.fileID = _tcstoul(token, NULL, 10);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		_tcscpy(node.user, token);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		_tcscpy(node.group, token);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		node.size = _tstol(token);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		node.speed = _tstol(token);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		node.type = *token;
		
		m_Log->dladd( (LPVOID) &node );
	}
	fclose(logfile);
	m_modified = false;

	return true;
}

// RD2 [fileID]=[user]|[group]|[mod]|[isChecked]|[isFreeFile]|
//
int AclCompare(const void *A, const void *B)
{
	double value = ((AclNode *)A)->fileID - ((AclNode *)B)->fileID;

	return ((value == 0) ? (0)
				: ((value < 0) ? (-1)
				: (1)));
}

bool CSimiyLog::WriteLogFile( )
{
	if( ! m_Log ) return false;

	if( FileExist( m_LogPath ) )
	{
		// Hidden File, can't write. why ? :(
		SetFileAttributes(m_LogPath, FILE_ATTRIBUTE_NORMAL);
	}

	if( m_Log->dlcount() == 0 )
	{
		_tunlink( m_LogPath );
		return true;
	}

	FILE *logfile = _tfopen( m_LogPath,  "wb");

	if( logfile == NULL)
	{
		SimiyIni.SfvChkLogAppend( "SimiyLog::WriteLogFile() fopen failed - %s", strerror(errno) );

		SetFileAttributes(m_LogPath, FILE_ATTRIBUTE_HIDDEN);	// back attribute
		return false;
	}
	// hidden log file
	SetFileAttributes( m_LogPath, FILE_ATTRIBUTE_HIDDEN );

	SNode *node;

	for( node = (SNode *) m_Log->dltofirst(); node; node = (SNode *)m_Log->dlgofwd() )
	{
		_ftprintf(logfile, "%.0f|%s|%s|%d|%d|%c\n",
			node->fileID, node->user, node->group, node->size, node->speed, node->type);
	}

	fclose(logfile);

	this->m_modified = false;
	return true;
}

void CSimiyLog::Add(LPCTSTR user,
		LPCTSTR group,
		long size,
		long speed,
		LPCTSTR filename,
		char type
		)
{
	if( ! m_Log ) 
	{
		SimiyIni.SfvChkLogAppend("SimiyLog::Add() Log is not readed - user = %s, group = %s, file = %s",
			user, group, filename);
		return;
	}

	if( SimiyIni.IsIncludeNoCntFExt( filename ) )	// skip file ext
		return;

	SNode node;

	node.fileID = getFileID(filename);
	_tcscpy(node.user, user);
	_tcscpy(node.group, group);
	node.size = size;
	node.speed = speed;
	node.type = type;

	m_Log->dladd( &node );
	m_modified = true;
}

void CSimiyLog::AddWithCheck(LPCTSTR user,
		LPCTSTR group,
		long size,
		long speed,
		LPCTSTR filename,
		char type
		)
{
	if( ! m_Log )
	{
		SimiyIni.SfvChkLogAppend("SimiyLog::AddWithCheck() Log is not readed - user = %s, group = %s, file = %s",
			user, group, filename);
		return;
	}

	if( SimiyIni.IsIncludeNoCntFExt( filename ) )	// skip file ext
		return;

	SNode node, *tmpNode;

	node.fileID = getFileID(filename);
	_tcscpy(node.user, user);
	_tcscpy(node.group, group);
	node.size = size;
	node.speed = speed;
	node.type = type;

	tmpNode = (SNode *) m_Log->dlfind( &node );

	if(tmpNode == NULL)
	{
		m_Log->dladd( &node );
		m_modified = true;
	}
	else if( tmpNode->type == UP_NDEF )
	{
		m_Log->dlreplace( &node );
		m_modified = true;
	}
}

int CSimiyLog::UserCountAndSpeedAndSize(LPCTSTR user,
		TCHAR *count,
		TCHAR *speed,
		TCHAR *size
		)
{
	if( ! m_Log )
	{
		SimiyIni.SfvChkLogAppend("SimiyLog::UserCountAndSpeedAndSize() Log is not readed - user = %s", user);
		if( count ) _tcscpy(count, "-1");
		if( size ) _tcscpy(size, "NULL");
		if( speed ) _tcscpy(speed, "NULL");
		return -1;
	}

	SNode *node;
	int ret_count = 0;
	double ret_size = 0;
	double ret_speed = 0;

	for( node = (SNode *) m_Log->dltofirst(); node; node = (SNode *)m_Log->dlgofwd() )
	{
		if( ! _tcscmp(node->user, user) )
		{
			ret_count ++;
			ret_size += node->size;
			if( speed )
			{
				// total time (sec)
				ret_speed += (double)(node->size) / (node->speed);
			}
		}
	}

	if( count ) _stprintf(count,"%d", ret_count);
	if( size ) _tcscpy(size, KByteToStr(ret_size));
	if( speed )
	{
		ret_speed = ret_size / ret_speed;
		_stprintf(speed, "%s%s", KByteToStr(ret_speed), "ps");
	}

	return ret_count;
}

SNode * CSimiyLog::GetNodeWithCheckBad(LPCTSTR upFilename)
{
	if( ! m_Log )
	{
		SimiyIni.SfvChkLogAppend("SimiyLog::GetNodeWithCheckBad() Log is not readed - file = %s", upFilename);
		return NULL;
	}
	int len = _tcslen( upFilename ) - 4;
	
	SNode *node = NULL;
	SNode searchNode;

	if( (len > 0) &&
		(_tcscmp(upFilename + len, ".bad") == 0) )	// this is bad-file format.
	{
		TCHAR path[_MAX_FNAME];

		_tcsncpy(path, upFilename, len);
		path[len] = _T('\0');

		searchNode.fileID = getFileID(path);

		node = (SNode *)m_Log->dlfind( &searchNode );
		if( !(node && node->type != 'S') )	// not found failed-file
		{
			node = NULL;
		}
	}
	if( ! node )
	{
		searchNode.fileID = getFileID(upFilename);
		node = (SNode *)m_Log->dlfind( &searchNode );
	}

	return node;
}

void CSimiyLog::DeleteNode(LPCTSTR file)
{
	SNode *node = GetNodeWithCheckBad( file );
	if( node )
	{
		m_Log->dldelete();
		m_modified = true;
	}
}

void CSimiyLog::DeleteNode(LPCTSTR filename,
		TCHAR *user,
		TCHAR *group
		)
{
	SNode *node = GetNodeWithCheckBad( filename );
	if( node )
	{
		if(user) _tcscpy(user, node->user);
		if(group) _tcscpy(group, node->group);
		m_Log->dldelete();
		m_modified = true;
	}
}

int RaceSizeCompare(const void *A, const void *B)
{
	return (
		((SNode *)B)->size - ((SNode *)A)->size	// b - a = rev sort
		);
}

int RaceCountCompare(const void *A, const void *B)
{
	return (
		((SNode *)B)->type - ((SNode *)A)->type
		);
}

int RaceUserCompare(const void *A, const void *B)
{
	return ( _tcscmp( ((SNode *)A)->user, ((SNode *)B)->user )
		);
}

int RaceGroupCompare(const void *A, const void *B)
{
	return ( _tcscmp( ((SNode *)A)->group, ((SNode *)B)->group )
		);
}

int SfvNodeCompare(const void *A, const void *B)
{
	double value = ((SfvNode *)A)->fileID - ((SfvNode *)B)->fileID;

	return ((value == 0) ? (0)
				: ((value < 0) ? (-1)
				: (1)));
}

void CSimiyLog::CalcRaceInfo( )
{
	if( ! m_Log )
	{
		SimiyIni.SfvChkLogAppend("SimiyLog::CalcRaceInfo() Log is not readed");
		return;
	}

	// RaceInfo initialize
	if( m_RaceUser )	delete m_RaceUser;
	if( m_RaceGroup )	delete m_RaceGroup;

	m_RacePeakFile = NULL;

	m_RaceUser = new CLinkedList( sizeof( SNode ), RaceUserCompare );
	m_RaceGroup = new CLinkedList( sizeof( SNode ), RaceGroupCompare );

	if( m_Sfv == NULL )
	{
		LoadSfvFile( );
	}

	this->m_RaceFilesSize = 0;
	double sfv_totalsecs = 0;

	SfvNode sfvNode;
	SNode *tmpNode;
	SNode addNode;
	for(SNode *node = (SNode *)m_Log->dltofirst();
		node;
		node = (SNode *)m_Log->dlgofwd() )
	{
		sfvNode.fileID = (unsigned long) node->fileID;

		if( m_Sfv->dlfind( &sfvNode ) )	// exist in sfv file list
		{
			// race user
			tmpNode = (SNode *) m_RaceUser->dlfind( node );
			if( tmpNode )
			{
				// exist, update
				tmpNode->size += node->size;
				tmpNode->fileID += (double)(node->size) / node->speed;
				tmpNode->type ++;
				
			}
			else
			{
				// first insert
				addNode = *node;
				addNode.fileID = (double)(node->size) / node->speed;
				addNode.type = 1;
				m_RaceUser->dladd( &addNode );
			}

			// race group
			tmpNode = (SNode *) m_RaceGroup->dlfind( node );
			if( tmpNode )
			{
				// exist, update
				tmpNode->size += node->size;
				tmpNode->fileID += (double)(node->size) / node->speed;
				tmpNode->type ++;
			}
			else
			{
				// first insert
				addNode = *node;
				addNode.fileID = (double)(node->size) / node->speed;
				addNode.type = 1;
				m_RaceGroup->dladd( &addNode );
			}

			this->m_RaceFilesSize += node->size;

			sfv_totalsecs += (double)(node->size) / node->speed;

			// Peak speed
			if( ! this->m_RacePeakFile ||
				node->speed > this->m_RacePeakFile->speed )
			{
				this->m_RacePeakFile = node;
			}
		}	// exist in sfv file list
	}
	this->m_RaceFilesCount = m_Sfv->dlcount();
	this->m_RaceAvgSpeed = (sfv_totalsecs == 0 ?
			0 : (int)(this->m_RaceFilesSize / sfv_totalsecs)
		);

	// basis is size

	this->m_RaceUser->dlsetcompare( RaceSizeCompare );
	this->m_RaceUser->dlqsort();

	this->m_RaceGroup->dlsetcompare( RaceSizeCompare );
	this->m_RaceGroup->dlqsort();

	// user, group speed
	for(tmpNode = (SNode *)m_RaceUser->dltofirst(); tmpNode; tmpNode = (SNode *)m_RaceUser->dlgofwd() )
	{
		tmpNode->speed = (int)(tmpNode->size / tmpNode->fileID);
	}
	for(tmpNode = (SNode *)m_RaceGroup->dltofirst(); tmpNode; tmpNode = (SNode *)m_RaceGroup->dlgofwd() )
	{
		tmpNode->speed = (int)(tmpNode->size / tmpNode->fileID);
	}

	return; // (ret_size);
}

void CSimiyLog::LoadSfvFile( )
{
	TCHAR sfv_file[_MAX_PATH];
	HANDLE findHandle;
	WIN32_FIND_DATA findData;

	if( m_Sfv )	return;
	m_Sfv = new CLinkedList( sizeof(SfvNode), SfvNodeCompare );

	_stprintf(sfv_file, "%s*.sfv", SFVCHK.UP_DIR);

	findHandle = FindFirstFile( sfv_file, &findData );
	if( findHandle != INVALID_HANDLE_VALUE )
	{
		GetFileName((LPCTSTR)findData.cFileName, sfv_file);
		FindClose(findHandle);
	}
	else
	{
		SimiyIni.SfvChkLogAppend( "SimiyLog::LoadSfvFile() sfv file is not exist (pattern = %s)", sfv_file );
		return;	//  ?
	}

	FILE *sfvfile = _tfopen(sfv_file, "rb");
	if(sfvfile == NULL) return; // 0;

	TCHAR buf[_MAX_PATH+16];	// 16 = space + crc_value
	int i;
	SfvNode sfvNode;

	// get sfv's file list
	while( _fgetts(buf, _MAX_PATH+16, sfvfile) )
	{
		if( FirstLChar( buf ) == _T(';') ) continue;	// comment

		LTrim( buf );
		RTrim( buf );

		for(i = _tcslen( buf ) - 1; i >= -1 && _istalnum( buf[i] ); i--);

		if( i < 0 ) continue; // filename is blank

		sfvNode.crc32 = _tcstoul( buf+i, NULL, 16);
		
		buf[i] = _T('\0');
		RTrim(buf);
		
		sfvNode.checked = false;
		sfvNode.fileID = getFileID( buf );
		_tcscpy(sfvNode.filename, buf);

		m_Sfv->dladd( &sfvNode );
	}
	fclose(sfvfile);
}

bool CSimiyLog::LoadAclFile(LPCTSTR UpDir )
{
	TCHAR buf[128];
	TCHAR path[_MAX_PATH];

	if( m_Acl ) return true;

	m_Acl = new CLinkedList( sizeof(AclNode), AclCompare);

	_stprintf(path, "%s.raidenftpd.acl", UpDir);

	if( ! FileExist(path) ) return true;

	FILE *aclfile = _tfopen( path,  "rb");
	
	if(aclfile == NULL)
	{
		SimiyIni.SfvChkLogAppend( "SimiyLog::LoadAclFile() fopen failed - %s", strerror(errno) );
		return false;
	}

	while( _fgetts(buf, 121, aclfile) )
	{
		AclNode node;

		if( _tcslen(buf) <= 16 ) continue;

		TCHAR *token = _tcstok( buf + 4, _TEXT("=") );
		if( token == NULL ) continue;
		node.fileID = _tcstoul(token, NULL, 10);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		_tcscpy(node.user, token);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		_tcscpy(node.group, token);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		_tcscpy(node.mod, token);

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		node.checked = *token;

		token = _tcstok(NULL, _TEXT("|"));
		if( token == NULL ) continue;
		node.freefile = *token;

		m_Acl->dladd( (LPVOID) &node );
	}
	fclose(aclfile);

	return true;
}

int CSimiyLog::GetSfvCount()
{
	if( m_SfvCount >= 0 ) return m_SfvCount;

	if( m_Sfv == NULL )
	{
		LoadSfvFile();
	}

	if( m_Acl == NULL )
	{
		LoadAclFile( SFVCHK.UP_DIR );
	}

	m_SfvCount = m_Sfv->dlcount();
	m_SfvGood = 0;
	m_SfvUpFileCnt = 0;

	for(SfvNode *node= (SfvNode *)m_Sfv->dltofirst(); node; node = (SfvNode *)m_Sfv->dlgofwd())
	{
		AclNode tmpNode;
		tmpNode.fileID = node->fileID;

		AclNode *aclNode = (AclNode *)m_Acl->dlfind( &tmpNode );	// this file exist in acl file

		if( aclNode && aclNode->checked == '1' )
		{
			node->checked = true;
			m_SfvGood ++;
		}

		SNode tmpSNode;
		tmpSNode.fileID = node->fileID;

		SNode *sNode = (SNode *)m_Log->dlfind( &tmpSNode );			// this file exist in log file
		if( sNode )
		{
			m_SfvUpFileCnt ++;
		}
	}

	return m_SfvCount;
}

int CSimiyLog::GetSfvGood()
{
	this->GetSfvCount();

	if( this->m_SfvUpFileCnt > 0 && this->m_SfvGood == 0 )
	{
		SfvCheckCrc32();
	}

	return m_SfvGood;
}

int CSimiyLog::SfvCheckCrc32()
{
	this->GetSfvCount();

	m_SfvGood = 0;	// re confirm

	int GENERAL_BAD = SimiyIni.GetValueInt("GENERAL","BAD", 1);
	TCHAR path[_MAX_PATH], tmpname[_MAX_PATH];

	// uploaded *.sfv in late, go check uploaded file
	for(SfvNode *node= (SfvNode *)m_Sfv->dltofirst(); node; node = (SfvNode *)m_Sfv->dlgofwd())
	{
		SNode tmpSNode;
		tmpSNode.fileID = node->fileID;

		SNode *sNode = (SNode *)m_Log->dlfind( &tmpSNode );			// this file exist in log file
		if( sNode )
		{
			_stprintf(path, "%s%s", SFVCHK.UP_DIR, node->filename);

			if( getFileCRC32( path ) == node->crc32 )
			{
				node->checked = true;
				this->m_SfvGood ++;
			}
			else
			{
				switch( GENERAL_BAD )
				{
				case 1:
					if( FileExist( path ) )	// not processed in raidenftpd
					{
						_stprintf(tmpname, "%s.bad", path);
						_tunlink( tmpname );					// delete pre-badfile
						_trename( path, tmpname );	// rename file to .bad
					}
					break;
				case 2:
					// delete up file & bad file
					_tunlink(path);
					_stprintf(tmpname, "%s.bad", path);
					_tunlink(tmpname);
					this->DeleteNode( node->filename );
				}
			}
		}
	}

	return m_SfvGood;
}

void CSimiyLog::MakeMissingFiles( )
{
	if( ! m_Log )
	{
		SimiyIni.SfvChkLogAppend("SimiyLog::MakeMissingFiles() Log is not readed");
		return;
	}

	TCHAR tmpname[_MAX_PATH];

	this->GetSfvGood();
	
	// all missing file is generated.
	for(SfvNode *node= (SfvNode *)m_Sfv->dltofirst(); node; node = (SfvNode *)m_Sfv->dlgofwd())
	{
		if( node->checked ) continue;	// this file is checked

		_stprintf(tmpname, "%s-missing", node->filename);

		// missing-file create
		FILE *fp = _tfopen(tmpname, "w");
		if( fp != NULL ) fclose(fp);
	}
}

const SNode *CSimiyLog::GetRaceUser(int ranking)
{
	if( !m_RaceUser || ranking <= 0 || ranking > this->m_RaceUser->dlcount() )
	{
		return NULL;
	}
	else
	{
		return (const SNode *) this->m_RaceUser->dlgetat(ranking - 1);
	}
}
const SNode *CSimiyLog::GetRaceGroup(int ranking)
{
	if( !m_RaceGroup || ranking <= 0 || ranking > this->m_RaceGroup->dlcount() )
	{
		return NULL;
	}
	else
	{
		return (const SNode *) this->m_RaceGroup->dlgetat(ranking - 1);
	}
}
int	CSimiyLog::GetRaceUsersCnt()
{
	return (int)(m_RaceUser ? this->m_RaceUser->dlcount() : 0);
}
int CSimiyLog::GetRaceGroupsCnt()
{
	return (int)(m_RaceGroup ? this->m_RaceGroup->dlcount() : 0);
}
int	CSimiyLog::GetRaceFilesCnt()
{
	return this->m_RaceFilesCount;
}
int CSimiyLog::GetRaceFilesSize()
{
	return this->m_RaceFilesSize;
}
int CSimiyLog::GetRaceAvgSpeed()
{
	return this->m_RaceAvgSpeed;
}
const SNode *CSimiyLog::GetRacePeakFile()
{
	return this->m_RacePeakFile;
}